
-------------------------------------------------------------------------------
--                                                                           --
--  Library Unit:  listing  --  Insert errors and messages into listing      --
--                                                                           --
--  Author:  Bradley L. Richards                                             --
--                                                                           --
--     Version     Date     Notes . . .                                      --
--    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -    --
--       1.0     6 Feb 86   Initial Version                                  --
--       1.1    25 Feb 86   Minor revisions to error messages                --
--       1.2     4 Mar 86   Added 2 character lookahead (required to         --
--                            differentiate between the Ada ellipse and      --
--                            a floating point number).                      --
--       1.3    22 May 86   Split message routines from io routines to       --
--                            limit visibility for higher routines           --
--       2.0    20 Jun 86   Added start and stop routines                    --
--       2.1    13 Jul 86   Split into separate spec and body files          --
--    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -    --
--                                                                           --
--  Library units used:  io                                                  --
--                                                                           --
--  Description:  This package inserts errors, warnings, and notes into      --
--     the listing file.  It tracks the numbers of messages of each type     --
--     which are inserted in the listing; these counts may be used, for      --
--     instance, to tell the user how many source errors were found.         --
--        The routines in this package all have a "pointer_flag" which       --
--     is used to tell the "io" package whether to draw a pointer to the     --
--     "current character" in the program being parsed.                      --
--                                                                           --
-------------------------------------------------------------------------------
--                                                                           --
--                              Package Body                                 --
--                                                                           --
-------------------------------------------------------------------------------

package body listing is

  procedure error( char_ptr : boolean; message : string ) is
    begin
      if char_ptr then
	print_pointer;
      end if;
      lput("Error -- "); lput_line(message);
      number_of_errors := number_of_errors + 1;
    end error;


  procedure note( char_ptr : boolean; message : string ) is
    begin
      if char_ptr then
	print_pointer;
      end if;
      lput("Note -- "); lput_line(message);
      number_of_notes := number_of_notes + 1;
    end note;


  procedure start_listing is
    begin
      number_of_errors := 0;
      number_of_notes := 0;
      number_of_warnings := 0;
    end start_listing;


  procedure stop_listing is
    begin
      null;
    end;


  procedure warning( char_ptr : boolean; message : string ) is
    begin
      if char_ptr then
	print_pointer;
      end if;
      lput("Warning -- "); lput_line(message);
      number_of_warnings := number_of_warnings + 1;
    end warning;

end listing;
